"""AST utilities."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
import hashlib
from typing import Dict, List
from sortedcontainers import SortedDict
from betterproto.lib.google.protobuf import Empty

from nada_mir_proto.nillion.nada.operations import v1 as proto_op
from nada_mir_proto.nillion.nada.types import v1 as proto_ty
from nada_mir_proto.nillion.nada.mir import v1 as proto_mir

from nada_dsl.nada_types import Party
from nada_dsl.source_ref import SourceRef


class OperationId:
    """Operation identifier generator."""

    current = 0

    @classmethod
    def next(cls):
        """Returns the next operation identifier."""
        next_op_id = cls.current
        cls.current += 1
        return next_op_id

    @classmethod
    def reset(cls):
        """Resets the operation identifier generator."""
        cls.current = 0


@dataclass
class ASTOperation(ABC):
    """AST Operations.

    Base abstract class for all AST Operations. Defines the minimum contract
    that needs to be implemented by any subclass.

    Attributes
    ----------
    id: int
        The operation identifier. This is the unique identifier assigned by Python
        that is the return value of id().

    source_ref: SourceRef
        The source reference object

    ty: str | Dict[str, Dict]
        The type representation, it can be a string (e.g. "SecretInteger") or a dictionary
        for compount types.
    """

    id: int
    source_ref: SourceRef
    ty: proto_ty.NadaType

    @abstractmethod
    def child_operations(self) -> List[int]:
        """Returns the list of identifiers of all the child operations of this operation."""
        raise NotImplementedError("Operation should implement child_operations method")

    @abstractmethod
    def to_mir(self) -> proto_op.Operation:
        """Converts this AST Operation into a valid MIR data structure"""
        raise NotImplementedError("Operation should implement to_mir method")


# Map of operations identified by the Python compiler
# The key is the operation identifier, the value the operation
AST_OPERATIONS: Dict[int, ASTOperation] = SortedDict()

# Map of literal hashes to index
LITERALS: Dict[str, int] = {}


@dataclass
class BinaryASTOperation(ASTOperation):
    """Superclass of all the Binary operations in AST representation"""

    variant: proto_op.BinaryOperationVariant
    left: int
    right: int

    def child_operations(self) -> List[int]:
        return [self.left, self.right]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            binary=proto_op.BinaryOperation(
                variant=self.variant,
                left=self.left,
                right=self.right,
            ),
        )


@dataclass
class UnaryASTOperation(ASTOperation):
    """Superclass of all the unary operations in AST representation"""

    variant: proto_op.UnaryOperationVariant
    child: int

    def child_operations(self):
        return [self.child]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            unary=proto_op.UnaryOperation(
                variant=self.variant,
                this=self.child,
            ),
        )


@dataclass
class IfElseASTOperation(ASTOperation):
    """AST Representation of an IfElse operation."""

    condition: int
    true_branch_child: int
    false_branch_child: int

    def child_operations(self):
        return [self.condition, self.true_branch_child, self.false_branch_child]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            ifelse=proto_op.IfElseOperation(
                cond=self.condition,
                first=self.true_branch_child,
                second=self.false_branch_child,
            ),
        )


@dataclass
class RandomASTOperation(ASTOperation):
    """AST Representation of a Random operation."""

    def child_operations(self):
        return []

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            random=Empty(),
        )


@dataclass
class InputASTOperation(ASTOperation):
    """AST representation of an Input."""

    name: str
    party: Party
    doc: str

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            input_ref=proto_op.InputReference(
                refers_to=self.name,
            ),
        )

    def child_operations(self) -> List[int]:
        return []


@dataclass
class LiteralASTOperation(ASTOperation):
    """AST Representation of a Literal."""

    name: str
    value: object
    literal_index: str

    def __init__(
        self,
        operation_id: int,
        name: str,
        ty: str | Dict[str, Dict],
        value: object,
        source_ref: SourceRef,
    ):
        self.id = operation_id
        self.name = name
        self.ty = ty
        self.value = value
        self.source_ref = source_ref
        # Generate a unique name depending on the value and type
        # to prevent duplicating literals in the bytecode.
        literal_name = hashlib.md5(
            (str(self.value) + str(self.ty)).encode("UTF-8")
        ).hexdigest()

        # Replace the literal name by an index so it is shorter
        if literal_name not in LITERALS:
            LITERALS[literal_name] = len(LITERALS)

        self.literal_index = str(LITERALS[literal_name])

        super().__init__(id=self.id, source_ref=self.source_ref, ty=self.ty)

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            literal_ref=proto_op.LiteralReference(
                refers_to=self.literal_index,
            ),
        )

    def child_operations(self) -> List[int]:
        return []


@dataclass
class ReduceASTOperation(ASTOperation):
    """AST Representation of a Reduce operation."""

    child: int
    fn: int
    initial: int

    def child_operations(self):
        return [self.child, self.initial]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            reduce=proto_op.ReduceOperation(
                fn=self.fn,
                child=self.child,
                initial=self.initial,
            ),
        )


@dataclass
class MapASTOperation(ASTOperation):
    """AST representation of a Map operation."""

    child: int
    fn: int

    def child_operations(self):
        return [self.child]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            map=proto_op.MapOperation(
                fn=self.fn,
                child=self.child,
            ),
        )


@dataclass
class NewASTOperation(ASTOperation):
    """AST Representation of a New operation."""

    name: str
    elements: List[int]

    def child_operations(self):
        return self.elements

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            new=proto_op.NewOperation(
                elements=self.elements,
            ),
        )


@dataclass
class NadaFunctionArgASTOperation(ASTOperation):
    """AST representation of a NadaFunctionArg operation."""

    name: str
    fn: int

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            arg_ref=proto_op.NadaFunctionArgRef(
                function_id=self.fn,
                refers_to=self.name,
            ),
        )

    def child_operations(self) -> List[int]:
        return []


@dataclass
class NadaFunctionASTOperation(ASTOperation):
    """AST representation of a nada function."""

    name: str
    args: List[int]
    child: int

    # pylint: disable=arguments-differ
    def to_mir(self, operations):
        """Convert a function to MIR."""
        args: List[proto_mir.NadaFunctionArg] = [
            proto_mir.NadaFunctionArg(
                name=AST_OPERATIONS[arg].name,
                type=AST_OPERATIONS[arg].ty,
                source_ref_index=AST_OPERATIONS[arg].source_ref.to_index(),
            )
            for arg in self.args
        ]  # type: ignore
        return proto_mir.NadaFunction(
            id=self.id,
            args=args,
            name=self.name,
            return_operation_id=self.child,
            operations=operations,
            return_type=self.ty,
            source_ref_index=self.source_ref.to_index(),
        )

    def __hash__(self) -> int:
        return self.id

    def child_operations(self) -> List[int]:
        return self.args + [self.child]


# Partially implemented
@dataclass
class CastASTOperation(ASTOperation):
    """AST Representation of a Cast operation."""

    target: int

    def child_operations(self):
        return [self.target]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            cast=proto_op.CastOperation(
                target=self.target,
                cast_to=self.ty,
            ),
        )


@dataclass
class TupleAccessorASTOperation(ASTOperation):
    """AST representation of a tuple accessor operation."""

    index: int
    source: int

    def child_operations(self):
        return [self.source]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            tuple_accessor=proto_op.TupleAccessor(
                index=proto_op.TupleIndex.LEFT
                if self.index == 0
                else proto_op.TupleIndex.RIGHT,
            ),
        )


@dataclass
class NTupleAccessorASTOperation(ASTOperation):
    """AST representation of a n tuple accessor operation."""

    index: int
    source: int

    def child_operations(self):
        return [self.source]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            ntuple_accessor=proto_op.NtupleAccessor(
                index=self.index,
                source=self.source,
            ),
        )


@dataclass
class ObjectAccessorASTOperation(ASTOperation):
    """AST representation of an object accessor operation."""

    key: str
    source: int

    def child_operations(self):
        return [self.source]

    def to_mir(self) -> proto_op.Operation:
        return proto_op.Operation(
            id=self.id,
            type=self.ty,
            source_ref_index=self.source_ref.to_index(),
            object_accessor=proto_op.ObjectAccessor(
                key=self.key,
                source=self.source,
            ),
        )



"""
Compiler frontend consisting of wrapper functions for the classes and functions
that constitute the Nada embedded domain-specific language (EDSL).
"""

from dataclasses import dataclass, field
import os
from typing import List, Dict, Tuple
from sortedcontainers import SortedDict


from nada_mir_proto.nillion.nada.mir import v1 as proto_mir
from nada_mir_proto.nillion.nada.operations import v1 as proto_op
from nada_mir_proto.nillion.nada.types import v1 as proto_ty


from nada_dsl import Party
from nada_dsl.ast_util import (
    AST_OPERATIONS,
    ASTOperation,
    BinaryASTOperation,
    CastASTOperation,
    IfElseASTOperation,
    InputASTOperation,
    LiteralASTOperation,
    MapASTOperation,
    TupleAccessorASTOperation,
    NTupleAccessorASTOperation,
    NadaFunctionASTOperation,
    NadaFunctionArgASTOperation,
    NewASTOperation,
    ObjectAccessorASTOperation,
    RandomASTOperation,
    ReduceASTOperation,
    UnaryASTOperation,
)
from nada_dsl.timer import timer
from nada_dsl.source_ref import SourceRef
from nada_dsl.program_io import Output


@dataclass
class CompilationContext:
    """
    Compilation context that holds the state of the compilation process.
    """

    inputs: Dict[Tuple[str, str], InputASTOperation] = field(default_factory=SortedDict)
    parties: Dict[str, Party] = field(default_factory=SortedDict)
    functions: Dict[int, NadaFunctionASTOperation] = field(default_factory=lambda: {})
    literals: Dict[str, Tuple[str, proto_ty.NadaType]] = field(
        default_factory=lambda: {}
    )


def get_target_dir() -> str:
    """Get the target directory for compilation output."""
    env_dir = os.environ.get("Nada_TARGET_DIR")
    if env_dir:
        return env_dir

    cwd = os.getcwd()

    try:
        os.mkdir("target")
    except FileExistsError:
        pass

    return os.path.join(cwd, "target")


def nada_compile(outputs: List[Output]) -> bytes:
    """Compile Nada to MIR and dump it as JSON."""
    compiled = nada_dsl_to_nada_mir(outputs)
    return bytes(compiled)


def nada_dsl_to_nada_mir(outputs: List[Output]) -> proto_mir.ProgramMir:
    """Convert Nada DSL to Nada MIR."""
    new_outputs = []
    ctx = CompilationContext()
    operations: Dict[int, proto_op.Operation] = SortedDict()
    # Process outputs
    for output in outputs:
        timer.start(
            f"nada_dsl.compiler_frontend.nada_dsl_to_nada_mir.{output.name}.process_operation"
        )
        out_operation_id = output.child.child.id
        traverse_and_process_operations(out_operation_id, operations, ctx)
        timer.stop(
            f"nada_dsl.compiler_frontend.nada_dsl_to_nada_mir.{output.name}.process_operation"
        )
        party = output.party
        ctx.parties[party.name] = party
        new_outputs.append(
            proto_mir.Output(
                operation_id=out_operation_id,
                name=output.name,
                party=party.name,
                type=AST_OPERATIONS[out_operation_id].ty,
                source_ref_index=output.source_ref.to_index(),
            )
        )

    operations = [
        proto_mir.OperationMapEntry(id=id, operation=op)
        for id, op in operations.items()
    ]

    mir = proto_mir.ProgramMir(
        functions=process_functions(ctx),
        parties=to_party_list(ctx.parties),
        inputs=to_input_list(ctx.inputs),
        literals=to_literal_list(ctx.literals),
        outputs=new_outputs,
        operations=operations,
        source_files=SourceRef.get_sources(),
        source_refs=SourceRef.get_refs(),
    )
    return mir


def to_party_list(parties: Dict[str, Party]) -> List[proto_mir.Party]:
    """Convert parties to a list in MIR format."""
    return [
        proto_mir.Party(
            name=party.name,
            source_ref_index=party.source_ref.to_index(),
        )
        for party in parties.values()
    ]


def to_input_list(inputs: Dict[int, InputASTOperation]) -> List[proto_mir.Input]:
    """Convert inputs to a list in MIR format."""
    input_list = []
    for input_ast in inputs.values():
        input_list.append(
            proto_mir.Input(
                name=input_ast.name,
                type=input_ast.ty,
                party=input_ast.party.name,
                doc=input_ast.doc,
                source_ref_index=input_ast.source_ref.to_index(),
            )
        )
    return input_list


def to_literal_list(
    literals: Dict[str, Tuple[str, proto_ty.NadaType]],
) -> List[proto_mir.Literal]:
    """Convert literals to a list in MIR format."""
    literal_list = []
    for name, (value, ty) in literals.items():
        literal_list.append(
            proto_mir.Literal(
                name=name,
                value=value,
                type=ty,
            )
        )
    return literal_list


def process_functions(
    ctx: CompilationContext,
) -> List[proto_mir.NadaFunction]:
    """Convert functions to a list in MIR format.

    From a starting dictionary of functions, it traverses each one of them,
    generating the corresponding MIR representation, discovering all the operations
    in the function.

    The algorithm might find new function calls while processing the operations
    in a function. These function calls might discover of new functions that are not
    in the original dictionary. These functions will be processed in turn.

    This function is designed to be invoked after the initial operation discovery
    which will find the starting set of functions.

    Arguments
    ---------
    functions: Dict[int, NadaFunctionASTOperation]
        A dictionary containing a starting list of functions
    """
    mir_functions = []
    stack = list(ctx.functions.values())
    ctx.functions = {}
    while len(stack) > 0:
        function = stack.pop()
        function_operations = SortedDict()

        traverse_and_process_operations(
            function.child,
            function_operations,
            ctx,
        )
        if len(ctx.functions) > 0:
            stack.extend(ctx.functions.values())
            ctx.functions = {}

        function_operations = [
            proto_mir.OperationMapEntry(id=id, operation=op)
            for id, op in function_operations.items()
        ]

        mir_functions.append(function.to_mir(function_operations))
    return mir_functions


def add_input_to_map(
    operation: InputASTOperation, ctx: CompilationContext
) -> proto_op.Operation:
    """Adds an input to the global INPUTS dictionary"""
    ctx.parties[operation.party.name] = operation.party
    if (operation.party.name, operation.name) in ctx.inputs and ctx.inputs[
        (operation.party.name, operation.name)
    ].id != operation.id:
        raise CompilerException(f"Input is duplicated: {operation.name}")

    ctx.inputs[(operation.party.name, operation.name)] = operation
    return operation.to_mir()


class CompilerException(Exception):
    """Generic compiler exception"""


def traverse_and_process_operations(
    operation_id: int,
    operations: Dict[int, proto_op.Operation],
    ctx: CompilationContext,
) -> Dict[int, NadaFunctionASTOperation]:
    """Traverses the AST operations finding all the operation tree rooted at the given
    operation. Uses an iterative DFS algorithm.

    It invokes `process_operation` which in turn generates a MIR and optionally discover
    extra functions.

    Arguments
    ---------
    operation_id: int
        The identifier of the root operation where the algorithm will start traversing the
        operation graph
    operations: Dict[int, Dict]
        Dictionary that will be updated with the operations found
    functions: Dict[int, NadaFunctionASTOperation]
        Dictionary of existing functions. If a function is found that is not in this dictionary
        it will added to the result dictionary

    Returns
    -------
    Dict[int, NadaFunctionASTOperation]
        Dictionary with all the new functions being found while traversing the operation tree
    """

    stack = [operation_id]
    while len(stack) > 0:
        operation_id = stack.pop()
        if operation_id not in operations:
            operation = AST_OPERATIONS[operation_id]
            maybe_op = process_operation(operation, ctx)
            if maybe_op is not None:
                operations[operation_id] = maybe_op
            stack.extend(operation.child_operations())


def process_operation(
    operation: ASTOperation, ctx: CompilationContext
) -> proto_op.Operation | None:
    """Process an AST operation.

    For arithmetic operations it simply returns the MIR representation of the operation.

    For inputs and literal types, it adds the corresponding value to the appropriate
    dictionaries and returns the MIR representation.

    For map, reduce and function call operations, adds the nada function if it's not in the
    functions dictionary, and returns the MIR representation.

    Whenever it finds a nada function, it adds it if it's not there. But it does not generate
    a MIR representation as functions are processed separately.

    It ignores nada function arguments as they should not be present in the MIR.
    """
    if isinstance(
        operation,
        (
            BinaryASTOperation,
            UnaryASTOperation,
            CastASTOperation,
            IfElseASTOperation,
            NewASTOperation,
            RandomASTOperation,
            NadaFunctionArgASTOperation,
            TupleAccessorASTOperation,
            NTupleAccessorASTOperation,
            ObjectAccessorASTOperation,
        ),
    ):
        return operation.to_mir()

    if isinstance(operation, InputASTOperation):
        add_input_to_map(operation, ctx)
        return operation.to_mir()
    if isinstance(operation, LiteralASTOperation):
        ctx.literals[operation.literal_index] = (str(operation.value), operation.ty)
        return operation.to_mir()
    if isinstance(operation, (MapASTOperation, ReduceASTOperation)):
        if operation.fn not in ctx.functions:
            ctx.functions[operation.fn] = AST_OPERATIONS[operation.fn]
        return operation.to_mir()
    if isinstance(operation, NadaFunctionASTOperation):
        if operation.id not in ctx.functions:
            ctx.functions[operation.id] = AST_OPERATIONS[operation.id]
        return None

    raise CompilerException(f"Compilation of Operation {operation} is not supported")


class PrintMirException(Exception):
    """Generic exception for printing MIR"""


def print_mir(mir: proto_mir.ProgramMir):
    """Prints the MIR in a human-readable format."""
    print("Parties:")
    for party in mir.parties:
        print(f"  {party.name}")
    print("Inputs:")
    for mir_input in mir.inputs:
        print(
            f"  {mir_input.name} ty({type_to_str(mir_input.type)}) party({mir_input.party})"
        )
    print("Literals:")
    for literal in mir.literals:
        print(f"  {literal.name} ty({type_to_str(literal.type)}) val({literal.value})")
    print("Outputs:")
    for output in mir.outputs:
        print(
            f"  {output.name} ty({type_to_str(output.type)}) oid({output.operation_id})"
        )
    print("Functions:")
    for function in mir.functions:
        args = ", ".join(
            [f"{arg.name}: ty({type_to_str(arg.type)})" for arg in function.args]
        )
        print(f"  {function.name} fn_id({function.id}), args({args})")
        print_operations(function.operations)

    print("Operations:")
    print_operations(mir.operations)


# pylint: disable=too-many-branches
def print_operations(operation: List[proto_mir.OperationMapEntry]):
    """Prints a list of operations in a human-readable format."""
    print()
    for entry in operation:
        op_id, op = entry.id, entry.operation
        line = f"oid({op_id}) rty({type_to_str(op.type)}) = "
        if hasattr(op, "binary"):
            line += f"{op.binary.variant} oid({op.binary.left}) oid({op.binary.right})"
        elif hasattr(op, "unary"):
            line += f"{op.unary.variant} oid({op.unary.this})"
        elif hasattr(op, "cast"):
            line += f"cast oid({op.cast.target})"
        elif hasattr(op, "ifelse"):
            line += f"ifelse cond({op.ifelse.cond}) true({op.ifelse.first}) false({op.ifelse.second})"
        elif hasattr(op, "random"):
            line += "random "
        elif hasattr(op, "input_ref"):
            line += f"input_ref to({op.input_ref.refers_to})"
        elif hasattr(op, "arg_ref"):
            line += (
                f"arg_ref fn_id({op.arg_ref.function_id}) to({op.arg_ref.refers_to})"
            )
        elif hasattr(op, "literal_ref"):
            line += f"literal_ref to({op.literal_ref.refers_to})"
        elif hasattr(op, "map"):
            line += f"map fn({(op.map.fn)}) oid({op.map.child})"
        elif hasattr(op, "reduce"):
            line += f"reduce fn({(op.reduce.fn)}) init({op.reduce.initial}) oid({op.reduce.child})"
        elif hasattr(op, "new"):
            oids = ", ".join([f"oid({oid})" for oid in op.new.elements])
            line += f"new {oids}"
        elif hasattr(op, "array_accessor"):
            line += f"array_accessor oid({op.array_accessor.source}) {op.array_accessor.index}"
        elif hasattr(op, "tuple_accessor"):
            line += f"tuple_accessor oid({op.tuple_accessor.source}) {op.tuple_accessor.index}"
        elif hasattr(op, "ntuple_accessor"):
            line += f"ntuple_accessor oid({op.ntuple_accessor.source}) {op.ntuple_accessor.index}"
        elif hasattr(op, "object_accessor"):
            line += f"object_accessor oid({op.object_accessor.source}) {op.object_accessor.key}"
        elif hasattr(op, "cast"):
            line += f"cast oid({op.cast.target}) {op.cast.to}"
        else:
            raise PrintMirException(f"Unknown operation {op}")
        print(line)


# pylint: disable=too-many-branches,too-many-return-statements
def type_to_str(ty: proto_ty.NadaType):
    """Converts a Nada type to a string."""
    if hasattr(ty, "integer"):
        return "Integer"
    if hasattr(ty, "unsigned_integer"):
        return "UnsignedInteger"
    if hasattr(ty, "boolean"):
        return "Boolean"
    if hasattr(ty, "secret_integer"):
        return "SecretInteger"
    if hasattr(ty, "secret_unsigned_integer"):
        return "SecretUnsignedInteger"
    if hasattr(ty, "secret_boolean"):
        return "SecretBoolean"
    if hasattr(ty, "ecdsa_private_key"):
        return "EcdsaPrivateKey"
    if hasattr(ty, "ecdsa_digest_message"):
        return "EcdsaDigestMessage"
    if hasattr(ty, "ecdsa_signature"):
        return "EcdsaSignature"
    if hasattr(ty, "array"):
        return f"Array[{type_to_str(ty.collection.contained_type)}:{ty.collection.array.size}]"
    if hasattr(ty, "tuple"):
        return f"Tuple[{type_to_str(ty.tuple.left)}, {type_to_str(ty.tuple.right)}]"
    if hasattr(ty, "object"):
        return "Object"
    if hasattr(ty, "ntuple"):
        return f"NTuple[{', '.join([type_to_str(t) for t in ty.ntuple.fields])}]"
    raise PrintMirException("Unknown type {ty}")



"""Operations that will be supported in the future."""

from dataclasses import dataclass

from nada_mir_proto.nillion.nada.types import v1 as proto_ty

from nada_dsl import SourceRef
from nada_dsl.ast_util import AST_OPERATIONS, CastASTOperation, OperationId
from nada_dsl.nada_types import AllTypes, AllTypesType


@dataclass
class Cast:
    """Cast operation."""

    target: AllTypes
    to: AllTypesType
    source_ref: SourceRef

    def __init__(self, target: AllTypes, to: AllTypes, source_ref: SourceRef):
        self.id = OperationId.next()
        self.target = target
        self.to = to
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST"""
        AST_OPERATIONS[self.id] = CastASTOperation(
            id=self.id, target=self.target, ty=ty, source_ref=self.source_ref
        )



"""
Compilation functions.
"""

import sys
import os.path
import base64
import json
from dataclasses import dataclass
import traceback
import importlib.util
from nada_dsl.compiler_frontend import nada_compile
from nada_dsl.errors import MissingEntryPointError, MissingProgramArgumentError
from nada_dsl.timer import add_timer, timer


@dataclass
class CompilerOutput:
    """Compiler Output"""

    mir: bytes


@add_timer(timer_name="nada_dsl.compile.compile")
def compile_script(script_path: str) -> CompilerOutput:
    """Compiles a NADA program

    Args:
        script_path (str): The nada program path

    Returns:
        CompilerOutput: The Compiler Output
    """
    script_dir = os.path.dirname(script_path)
    sys.path.insert(0, script_dir)
    script_name = os.path.basename(script_path)
    if script_name.endswith(".py"):
        script_name = script_name[:-3]
    timer.start("nada_dsl.compile.compile.__import__")
    script = __import__(script_name)
    timer.stop("nada_dsl.compile.compile.__import__")

    try:
        main = getattr(script, "nada_main")
    except Exception as exc:
        raise MissingEntryPointError(
            "'nada_dsl' entrypoint function is missing in program " + script_name
        ) from exc
    outputs = main()
    compile_output = nada_compile(outputs)
    return CompilerOutput(compile_output)


@add_timer(timer_name="nada_dsl.compile.compile_string")
def compile_string(script: str) -> CompilerOutput:
    """Compiles a NADA program from a string

    Args:
        script (str): The nada program as a base64 encoded string (UTF-8)

    Returns:
        CompilerOutput: The Compiler Output
    """
    decoded_program = base64.b64decode(script).decode("utf-8")
    temp_name = "temp_program"
    spec = importlib.util.spec_from_loader(temp_name, loader=None)
    module = importlib.util.module_from_spec(spec)
    exec(decoded_program, module.__dict__)  # pylint:disable=W0122
    sys.modules[temp_name] = module
    globals()[temp_name] = module

    outputs = module.nada_main()
    compile_output = nada_compile(outputs)
    return CompilerOutput(compile_output)


def print_output(out: CompilerOutput):
    """Prints compiler output

    Args:
        out (CompilerOutput): Output of the compiler
    """
    output_json = {
        "result": "Success",
        "mir": list(out.mir),
    }
    print(json.dumps(output_json))


if __name__ == "__main__":
    try:
        if os.environ.get("NADA_TIMER"):
            timer.enable()
        args_length = len(sys.argv)
        if args_length < 2:
            raise MissingProgramArgumentError("expected program as argument")
        if args_length == 2:
            output = compile_script(sys.argv[1])
            print_output(output)
        if args_length == 3 and sys.argv[1] == "-s":
            output = compile_string(sys.argv[2])
            print_output(output)

    except Exception as ex:
        output = {
            "result": "Failure",
            "reason": str(ex),
            "traceback": str(traceback.format_exc()),
        }
        print(json.dumps(output))

    finally:
        if timer.is_enabled():
            with open("nada-timers.json", "w", encoding="utf-8") as fd:
                json.dump(timer.report(), fd)



"""
Nada DSL Exceptions.
"""


class NotAllowedException(Exception):
    """Exception for not allowed use cases."""


class InvalidTypeError(Exception):
    """Invalid type error"""


class MissingProgramArgumentError(Exception):
    """Missing program argument"""


class MissingEntryPointError(Exception):
    """Missing nada_main entry point for the program."""


class IncompatibleTypesError(Exception):
    """The types in an operation are not compatible."""



"""
Common functions that Nada DSL static analysis submodules can use to build
interactive HTML reports.
"""

# pylint: disable=wildcard-import,unused-wildcard-import,invalid-name
from __future__ import annotations
from typing import List, Tuple
import ast
import asttokens
import richreports
import parsial

try:
    from nada_dsl.audit.abstract import *
    from nada_dsl.audit.common import *
except:  # pylint: disable=bare-except # For Nada DSL Sandbox support.
    from abstract import *
    from common import *


def parse(source: str) -> Tuple[asttokens.ASTTokens, List[int]]:
    """
    Parse a Python source string that represents a Nada DSL program
    and return both its abstract syntax tree and a list of which lines
    were skipped by the partial parser due to syntax errors.
    """
    lines = source.split("\n")
    (_, slices) = parsial.parsial(ast.parse)(source)
    lines_ = [l[s] for (l, s) in zip(lines, slices)]
    skips = [i for i in range(len(lines)) if len(lines[i]) != len(lines_[i])]
    return (asttokens.ASTTokens("\n".join(lines_), parse=True), skips)


def locations(report_, asttokens_, a):
    """
    Return the starting and ending locations corresponding to an :obj:`ast`
    node.
    """
    ((start_line, start_column), (end_line, end_column)) = (
        asttokens_.get_text_positions(a, True)
    )

    # Skip any whitespace when determining the starting location.
    line = report_.lines[start_line - 1]
    while line[start_column] == " " and start_column < len(line):
        start_column += 1

    return (
        richreports.location((start_line, start_column)),
        richreports.location((end_line, end_column - 1)),
    )


def type_to_str(t):
    """
    Convert a type and/or type error :obj:`ast` node attribute into a
    human-readable string.
    """
    if hasattr(t, "__name__"):
        if t.__name__ == "list":
            if hasattr(t, "__args__"):
                return "list[" + type_to_str(t.__args__[0]) + "]"
            return "list"

        return str(t.__name__)

    if isinstance(t, TypeError):
        return str("TypeError: " + str(t))

    return str("TypeError: " + "type cannot be determined")


def enrich_from_type(report_, type_, start, end):
    """
    Enrich a range within a report according to the supplied type attribute.
    """
    if type_ in (
        bool,
        int,
        str,
        range,
        Party,
        Input,
        Output,
        Integer,
        PublicInteger,
        SecretInteger,
        Boolean,
        PublicBoolean,
        SecretBoolean,
    ) or (hasattr(type_, "__name__") and type_.__name__ == "list"):
        t_str = type_to_str(type_)
        report_.enrich(
            start, end, '<span class="types-' + t_str + '">', "</span>", True, True
        )

    if isinstance(type_, (TypeError, TypeErrorRoot)):
        report_.enrich(
            start,
            end,
            '<span class="types-' + type_.__class__.__name__ + '">',
            "</span>",
            True,
            True,
        )


def enrich_syntaxrestriction(report_, r, start, end):
    """
    Enrich a range within a report according to the supplied syntax restriction.
    """
    report_.enrich(
        start,
        end,
        '<span class="rules-SyntaxRestriction">',
        "</span>",
        enrich_intermediate_lines=True,
        skip_whitespace=True,
    )
    report_.enrich(
        start,
        end,
        '<span class="detail" data-detail="SyntaxRestriction: ' + str(r) + '">',
        "</span>",
        enrich_intermediate_lines=True,
        skip_whitespace=True,
    )


def enrich_keyword(report_, start, length):
    """
    Enrich a range within a report corresponding to a Python keyword.
    """
    (start_line, start_column) = start
    report_.enrich(
        (start_line, start_column),
        (start_line, start_column + length),
        '<span class="keyword">',
        "</span>",
        enrich_intermediate_lines=True,
        skip_whitespace=True,
    )


def enrich_fromaudits(report_: richreports.report, atok) -> richreports.report:
    """
    Enrich a report containing the source code of a Nada DSL program using the
    static analysis attributes found within an :obj:`ast` instance.
    """
    # pylint: disable=too-many-statements,too-many-branches
    for a in ast.walk(atok.tree):
        r = audits(a, "rules")
        t = audits(a, "types")

        if isinstance(a, (ast.Assign, ast.AnnAssign)):
            target = a.targets[0] if hasattr(a, "targets") else a.target
            (start, end) = locations(report_, atok, target)
            if isinstance(r, SyntaxRestriction):
                enrich_syntaxrestriction(report_, r, start, end)
            else:
                enrich_from_type(report_, t, start, end)
                t_str = (
                    type_to_str(t)
                    if not isinstance(t, TypeError)
                    else "TypeError: " + str(t)
                )
                report_.enrich(
                    start,
                    end,
                    '<span class="detail" data-detail="' + t_str + '">',
                    "</span>",
                    True,
                )

        elif isinstance(r, SyntaxRestriction):
            (start, end) = locations(report_, atok, a)
            enrich_syntaxrestriction(report_, r, start, end)

        elif isinstance(r, RuleInAncestor):
            pass  # This node will be wrapped by an ancestor's enrichment.

        elif isinstance(a, ast.ImportFrom):
            (start, _) = locations(report_, atok, a)
            enrich_keyword(report_, start, 4)

        elif isinstance(a, ast.Return):
            (start, end) = locations(report_, atok, a)
            if isinstance(r, SyntaxRestriction):
                enrich_syntaxrestriction(report_, r, start, end)
            else:
                enrich_keyword(report_, start, 6)
                t = audits(a.value, "types")
                report_.enrich(
                    start,
                    start + (0, 6),
                    '<span class="detail" data-detail="' + type_to_str(t) + '">',
                    "</span>",
                    True,
                )

        elif isinstance(a, ast.For):
            (start, end) = locations(report_, atok, a)
            enrich_keyword(report_, start, 3)
            (_, start) = locations(report_, atok, a.target)
            (end, _) = locations(report_, atok, a.iter)
            report_.enrich(
                start + (0, 1),
                end - (0, 1),
                '<span class="keyword">',
                "</span>",
                enrich_intermediate_lines=True,
                skip_whitespace=True,
            )

        elif isinstance(a, ast.FunctionDef):
            (start, end) = locations(report_, atok, a)
            enrich_keyword(report_, start, 3)
            if isinstance(r, SyntaxRestriction):
                enrich_syntaxrestriction(report_, r, start, end)

        elif isinstance(a, ast.ListComp):
            for generator in a.generators:
                (start, _) = locations(report_, atok, generator)
                enrich_keyword(report_, start, 3)
                (_, start) = locations(report_, atok, generator.target)
                (end, _) = locations(report_, atok, generator.iter)
                report_.enrich(
                    start + (0, 1),
                    end - (0, 1),
                    '<span class="keyword">',
                    "</span>",
                    enrich_intermediate_lines=True,
                    skip_whitespace=True,
                )

        elif isinstance(a, ast.Call):
            (start, end) = locations(report_, atok, a.func)
            if isinstance(a.func, ast.Attribute):
                (_, start) = locations(report_, atok, a.func.value)
                start = start + (0, 1)
            enrich_from_type(report_, t, start, end)
            report_.enrich(
                start,
                end,
                '<span class="detail" data-detail="' + type_to_str(t) + '">',
                "</span>",
                True,
            )

        elif isinstance(a, ast.BoolOp):
            for i in range(len(a.values) - 1):
                (left, right) = (a.values[i], a.values[i + 1])
                (_, end) = locations(report_, atok, left)
                (start, _) = locations(report_, atok, right)
                (start, end) = (end + (0, 1), start - (0, 1))
                report_.enrich(start, end, "<b>", "</b>", True, True)
                enrich_from_type(report_, t, start, end)
                report_.enrich(
                    start,
                    end,
                    '<span class="detail" data-detail="' + type_to_str(t) + '">',
                    "</span>",
                    True,
                    True,
                )

        elif isinstance(a, ast.BinOp):
            (_, end) = locations(report_, atok, a.left)
            (start, _) = locations(report_, atok, a.right)
            (start, end) = (end + (0, 1), start - (0, 1))
            enrich_from_type(report_, t, start, end)
            report_.enrich(
                start,
                end,
                '<span class="detail" data-detail="' + type_to_str(t) + '">',
                "</span>",
                True,
                True,
            )

        elif isinstance(a, ast.Compare):
            (_, end) = locations(report_, atok, a.left)
            (start, _) = locations(report_, atok, a.comparators[0])
            (start, end) = (end + (0, 1), start - (0, 1))
            enrich_from_type(report_, t, start, end)
            report_.enrich(
                start,
                end,
                '<span class="detail" data-detail="' + type_to_str(t) + '">',
                "</span>",
                True,
                True,
            )

        elif isinstance(a, ast.UnaryOp):
            (start, _) = locations(report_, atok, a)
            (end, _) = locations(report_, atok, a.operand)
            end = end - (0, 2)
            enrich_from_type(report_, t, start, end)
            if isinstance(a.op, ast.Not):
                report_.enrich(start, end, "<b>", "</b>", True)
            report_.enrich(
                start,
                end,
                '<span class="detail" data-detail="' + type_to_str(t) + '">',
                "</span>",
                True,
            )

        elif isinstance(a, ast.Constant):
            (start, end) = locations(report_, atok, a)
            if isinstance(r, SyntaxRestriction):
                enrich_syntaxrestriction(report_, r, start, end)
            else:
                enrich_from_type(report_, t, start, end)
                report_.enrich(
                    start,
                    end,
                    '<span class="detail" data-detail="' + type_to_str(t) + '">',
                    "</span>",
                    True,
                )

        elif isinstance(a, ast.Name):
            (start, end) = locations(report_, atok, a)
            if isinstance(r, SyntaxRestriction):
                enrich_syntaxrestriction(report_, r, start, end)
            else:
                if t is not None and not isinstance(t, TypeInParent):
                    enrich_from_type(report_, t, start, end)
                    report_.enrich(
                        start,
                        end,
                        '<span class="detail" data-detail="' + type_to_str(t) + '">',
                        "</span>",
                        True,
                    )


def html(report: richreports.report) -> str:
    """
    Return a self-contained CSS/HTML document corresponding to a report.
    """
    head = (
        "    "
        + """
    <style>
      body { font-family:Monospace; white-space;pre; }
      .keyword { font-weight:bold; color:#000000; }
      .rules-SyntaxError { background-color:#FFFF00; color:#555555; }
      .rules-SyntaxRestriction { background-color:#CCCC00; font-weight:bold; color:#000000; }
      .types-TypeError { background-color:#FFCCCB; font-weight:bold; color:#B22222; }
      .types-TypeErrorRoot { background-color:#FF0000; font-weight:bold; color:#FFFFFF ; }
      .types-bool { font-weight:bold; color:#000000; }
      .types-int { font-weight:bold; color:#000000; }
      .types-str { font-weight:bold; color:#000000; }
      .types-Party { font-weight:bold; color:#d45cd6; }
      .types-Input { font-weight:bold; color:#DAA520; }
      .types-Output { font-weight:bold; color:#DAA520; }
      .types-Integer { font-weight:bold; color:#009900; }
      .types-Integer { font-weight:bold; color:#009900; }
      .types-PublicInteger { font-weight:bold; color:#009900; }
      .types-SecretInteger { font-weight:bold; color:#0000FF; }
      .types-Boolean { font-weight:bold; color:#009900; }
      .types-PublicBoolean { font-weight:bold; color:#009900; }
      .types-SecretBoolean { font-weight:bold; color:#0000FF; }
      div { height:18px; line-height:18px; white-space:pre; }
      div span { padding-top:3px; padding-bottom:3px; line-height:18px; cursor:pointer; }
      .detail { cursor:pointer; }
      .detail:hover { background-color:#DDDDDD; }
      #detail {
        display:none;
        position:absolute;
        padding:12px;
        background-color:#000000;
        color:#FFFFFF;
      }
    </style>
    """.strip()
        + "\n"
    )

    script = (
        "    "
        + """
    <script>
      window.onload = function () {
        const elements = document.getElementsByClassName("detail");
        for (let i = 0; i < elements.length; i++) {
          const element = elements[i];
          element.addEventListener("mouseover", function (event) {
            const detail = document.getElementById("detail");
            detail.style.display = "block";
            detail.style.top = element.offsetTop + 24 + "px";
            detail.style.left = (
              element.offsetLeft +
              element.getBoundingClientRect().width -
              16
            ) + "px";
            detail.innerHTML = element.dataset.detail;
          });
          element.addEventListener("mouseout", function (event) { 
            const detail = document.getElementById("detail");
            detail.style.display = "none";
          });
        }
      };
    </script>
    """.strip()
        + "\n"
    )

    return (
        "<html>\n"
        + "  <head>\n"
        + head
        + "  </head>\n"
        + '  <body>\n  <div id="detail"></div>\n'
        + report.render()
        + "\n  </body>\n"
        + script
        + "</html>\n"
    )



"""function.py

Nada function definitions and utilities.
"""

import inspect
from typing import Generic, List, Callable

from nada_mir_proto.nillion.nada.types import v1 as proto_ty

from nada_dsl import SourceRef
from nada_dsl.ast_util import (
    AST_OPERATIONS,
    NadaFunctionASTOperation,
    NadaFunctionArgASTOperation,
    OperationId,
)
from nada_dsl.nada_types.generics import T, R
from nada_dsl.nada_types import DslType


class NadaFunctionArg(Generic[T]):
    """Represents a Nada function argument."""

    function_id: int
    name: str
    type: T
    source_ref: SourceRef

    def __init__(self, function_id: int, name: str, arg_type: T, source_ref: SourceRef):
        self.id = OperationId.next()
        self.function_id = function_id
        self.name = name
        self.type = arg_type
        self.source_ref = source_ref
        self.store_in_ast(arg_type.to_mir())

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST."""
        AST_OPERATIONS[self.id] = NadaFunctionArgASTOperation(
            id=self.id,
            name=self.name,
            fn=self.function_id,
            ty=ty,
            source_ref=self.source_ref,
        )


class NadaFunction(Generic[T, R]):
    """Nada Function.

    Represents a Nada Function. Nada functions are special types of functions that are used
    in map / reduce operations.
    """

    id: int
    args: List[NadaFunctionArg]
    function: Callable[[T], R]
    return_type: R
    source_ref: SourceRef

    def __init__(
        self,
        function_id: int,
        args: List[NadaFunctionArg],
        function: Callable[[T], R],
        return_type: R,
        source_ref: SourceRef,
        child: DslType,
    ):
        self.child = child
        self.id = function_id
        self.args = args
        self.function = function

        self.return_type = return_type
        self.source_ref = source_ref
        self.store_in_ast()

    def store_in_ast(self):
        """Store this Nada Function in AST."""
        AST_OPERATIONS[self.id] = NadaFunctionASTOperation(
            name=self.function.__name__,
            args=[arg.id for arg in self.args],
            id=self.id,
            ty=self.return_type.to_mir(),
            source_ref=self.source_ref,
            child=self.child.child.id,
        )


def create_nada_fn(fn, args_ty) -> NadaFunction[T, R]:
    """
    Can be used also for lambdas
    ```python
    array.map(
        nada_fn(
            lambda x: x.cast(SecretInteger),
            args_ty={'x': SecretInteger}, return_ty=SecretInteger))
    ```
    """

    args = inspect.getfullargspec(fn)
    nada_args = []
    function_id = OperationId.next()
    nada_args_type_wrapped = []
    for arg, arg_ty in zip(args.args, args_ty):
        # We'll get the function source ref for now
        nada_arg = NadaFunctionArg(
            function_id,
            name=arg,
            arg_type=arg_ty,
            source_ref=SourceRef.back_frame(),
        )
        nada_args.append(nada_arg)
        nada_args_type_wrapped.append(arg_ty.instantiate(nada_arg))

    child = fn(*nada_args_type_wrapped)

    return_type = child.type()
    return NadaFunction(
        function_id,
        function=fn,
        args=nada_args,
        child=child,
        return_type=return_type,
        source_ref=SourceRef.back_frame(),
    )



"""
Static analysis submodule that defines the highly limited "strict"
subset of the Nada DSL syntax.
"""

# pylint: disable=wildcard-import,unused-wildcard-import,invalid-name
from __future__ import annotations
from typing import Callable
import ast
import richreports

try:
    from nada_dsl.audit.abstract import *
    from nada_dsl.audit.common import *
    from nada_dsl.audit.report import parse, enrich_fromaudits
except:  # pylint: disable=bare-except # For Nada DSL Sandbox support.
    from abstract import *
    from common import *
    from report import parse, enrich_fromaudits


def _rules_restrictions_descendants(a):
    """
    Remove all rule attributes that are redundant because they occur in a
    an ancestor.
    """
    for a_ in ast.walk(a):
        if not isinstance(a_, (ast.Assign, ast.Assign)):
            if isinstance(audits(a_, "rules"), (SyntaxRestriction, RuleInAncestor)):
                for a__ in ast.walk(a_):
                    if a__ != a_:
                        audits(a__, "rules", RuleInAncestor())
                        audits(a__, "types", delete=True)


def rules(a):
    """
    Mark all :obj:`ast` nodes as prohibited (as a starting point for a strict
    static analysis).
    """
    if isinstance(a, ast.Module):
        for a_ in ast.walk(a):
            audits(
                a_,
                "rules",
                SyntaxRestriction("use of this syntax is prohibited in strict mode"),
            )


def _types_base(t):
    """
    Return a boolean value indicating whether the supplied type is a base
    type.
    """
    return t in (
        bool,
        int,
        str,
        Integer,
        PublicInteger,
        SecretInteger,
        Boolean,
        PublicBoolean,
        SecretBoolean,
    )


def _types_list_monomorphic(t):
    """
    Return a boolean value indicating whether the supplied type represents a
    monomorphic list type (*i.e.*, a list type wherein the types of the items
    are fully specified).
    """
    if t.__name__ == "list" and hasattr(t, "__args__") and len(t.__args__) == 1:
        return _types_list_monomorphic(t.__args__[0])

    if _types_base(t):
        return True

    return False


def _types_list_monomorphic_depth(t):
    """
    Return an integer representing the depth of the monomorphic list type.
    """
    if t.__name__ == "list" and hasattr(t, "__args__") and len(t.__args__) == 1:
        return 1 + _types_list_monomorphic_depth(t.__args__[0])

    return 0


def _types_monomorphic(t):
    """
    Return a boolean value indicating whether the supplied type represents a
    monomorphic type permitted by this static analysis submodule.
    """
    return _types_base(t) or _types_list_monomorphic(t)


def _types_binop_mult_add_sub(t_l, t_r):
    """
    Determine the result type for multiplication and addition operations
    involving Nada DSL integers.
    """
    t = TypeErrorRoot("arguments must have integer types")
    if (t_l, t_r) == (Integer, Integer):
        t = Integer
    elif (t_l, t_r) == (Integer, PublicInteger):
        t = PublicInteger
    elif (t_l, t_r) == (PublicInteger, Integer):
        t = PublicInteger
    elif (t_l, t_r) == (Integer, SecretInteger):
        t = SecretInteger
    elif (t_l, t_r) == (SecretInteger, Integer):
        t = SecretInteger
    elif (t_l, t_r) == (PublicInteger, PublicInteger):
        t = PublicInteger
    elif (t_l, t_r) == (PublicInteger, SecretInteger):
        t = SecretInteger
    elif (t_l, t_r) == (SecretInteger, PublicInteger):
        t = SecretInteger
    elif (t_l, t_r) == (SecretInteger, SecretInteger):
        t = SecretInteger

    if isinstance(t_l, TypeError):
        t = typeerror_demote(t_l)
    elif isinstance(t_r, TypeError):
        t = typeerror_demote(t_r)

    return t


def _types_compare(t_l, t_r):
    """
    Determine the result type for comparison operations involving Nada DSL
    integers.
    """
    t = TypeErrorRoot("arguments must have integer types")
    if (t_l, t_r) == (Integer, Integer):
        t = Boolean
    elif (t_l, t_r) == (Integer, PublicInteger):
        t = PublicBoolean
    elif (t_l, t_r) == (PublicInteger, Integer):
        t = PublicBoolean
    elif (t_l, t_r) == (Integer, SecretInteger):
        t = SecretBoolean
    elif (t_l, t_r) == (SecretInteger, Integer):
        t = SecretBoolean
    elif (t_l, t_r) == (PublicInteger, PublicInteger):
        t = PublicBoolean
    elif (t_l, t_r) == (PublicInteger, SecretInteger):
        t = SecretBoolean
    elif (t_l, t_r) == (SecretInteger, PublicInteger):
        t = SecretBoolean
    elif (t_l, t_r) == (SecretInteger, SecretInteger):
        t = SecretBoolean

    if isinstance(t_l, TypeError):
        t = typeerror_demote(t_l)
    elif isinstance(t_r, TypeError):
        t = typeerror_demote(t_r)

    return t


def types(a, env=None, func=False):
    """
    Infer types of :obj:`ast` where possible, adding the type (or error)
    information in node attributes (and removing syntax restriction attributes
    as appropriate).
    """
    # pylint: disable=too-many-statements,too-many-return-statements
    # pylint: disable=too-many-nested-blocks,too-many-branches
    # pylint: disable=too-many-locals
    env = {} if env is None else env

    # Handle cases in which the input node is a statement.
    if isinstance(a, ast.Module):
        rules_no_restriction(a)
        for a_ in a.body:
            env = types(a_, env, func)

        # Remove syntax restriction attributes that appear for descendants
        # of nodes that already have such attributes.
        _rules_restrictions_descendants(a)

        return env

    if isinstance(a, ast.ImportFrom):
        if (
            a.module == "nada_dsl"
            and len(a.names) == 1
            and a.names[0].name == "*"
            and a.names[0].asname is None
            and a.level == 0
        ):
            rules_no_restriction(a, recursive=True)
        return env

    if isinstance(a, ast.FunctionDef):
        if not func:
            if a.name == "nada_main":
                rules_no_restriction(a)
                rules_no_restriction(a.args)

                # Create a local copy of the environment. Only the
                # original environment passed to this invocation
                # is returned.
                env_ = dict(env)
                for a_ in a.body:
                    env_ = types(a_, env_, func=True)
            else:
                rules_no_restriction(a)
                rules_no_restriction(a.args)

                t_ret = eval(ast.unparse(a.returns))  # pylint: disable=eval-used
                if _types_monomorphic(t_ret):
                    rules_no_restriction(a.returns)

                env_ = dict(env)
                ts = []
                for arg in a.args.args:
                    var = arg.arg
                    t_var = eval(ast.unparse(arg.annotation))  # pylint: disable=eval-used
                    if _types_monomorphic(t_var):
                        rules_no_restriction(arg)
                        rules_no_restriction(arg.annotation, recursive=True)
                        env_[var] = t_var
                        ts.append(t_var)
                for a_ in a.body:
                    env_ = types(a_, env_, func=True)
                env[a.name] = Callable[ts, t_ret]

        return env

    if isinstance(a, ast.Assign):
        if len(a.targets) == 1:
            target = a.targets[0]
            if isinstance(target, ast.Name):
                rules_no_restriction(a)
                rules_no_restriction(target)
                types(a.value, env, func)
                t = audits(a.value, "types")
                if t == list and not _types_list_monomorphic(t):
                    t = TypeErrorRoot(
                        "assignment of list value with underspecified "
                        + "type requires fully specified type annotation"
                    )
                    audits(a, "types", t)
                elif t is not None:
                    audits(a, "types", typeerror_demote(t))
                    audits(target, "types", TypeInParent())
                    if not isinstance(t, TypeError):
                        if isinstance(target, ast.Name):
                            var = target.id
                            env[var] = t
            elif isinstance(target, ast.Subscript):
                rules_no_restriction(a)
                types(a.value, env, func)
                t = audits(a.value, "types")

                target_ = target
                invalid_index = False
                depth = 0
                while isinstance(target_, ast.Subscript):
                    depth += 1
                    rules_no_restriction(target_)
                    types(target_.slice, env, func)
                    t_s = audits(target_.slice, "types")
                    if t_s != int:
                        invalid_index = True
                        break
                    if isinstance(target_.value, (ast.Name, ast.Subscript)):
                        target_ = target_.value

                if invalid_index:
                    audits(a, "types", TypeErrorRoot("indices must be integers"))

                if isinstance(target_, ast.Name):
                    rules_no_restriction(target)
                    rules_no_restriction(target_)
                    if target_.id in env:
                        t_b = env[target_.id]
                        if _types_list_monomorphic_depth(t_b) < depth:
                            audits(
                                a,
                                "types",
                                TypeErrorRoot("target has incompatible type"),
                            )
                        else:
                            audits(a, "types", typeerror_demote(t))
                    else:
                        audits(
                            a, "types", TypeErrorRoot("unbound variable: " + target_.id)
                        )

        return env

    if isinstance(a, ast.AnnAssign):
        if isinstance(a.target, ast.Name):
            rules_no_restriction(a)
            rules_no_restriction(a.target)
            types(a.value, env, func)
            t = audits(a.value, "types")
            try:
                t_a = eval(ast.unparse(a.annotation))  # pylint: disable=eval-used
                rules_no_restriction(a.annotation, recursive=True)
                if not _types_list_monomorphic(t_a):
                    audits(
                        a.annotation,
                        "types",
                        TypeErrorRoot(
                            "assignment of list value requires fully specified type annotation"
                        ),
                    )
            except:  # pylint: disable=bare-except
                t_a = TypeErrorRoot("invalid type annotation")

            if isinstance(t, TypeError):
                audits(a, "types", t)
                audits(a.target, "types", t)
            else:
                t_u = unify(t_a, t)
                if t_u is None:
                    t = TypeErrorRoot(
                        "value type cannot be reconciled with type annotation"
                    )
                    audits(a, "types", t)
                    audits(a.target, "types", t)
                else:
                    t = t_u
                    audits(a, "types", t)
                    audits(a.target, "types", TypeInParent())
                    var = a.target.id
                    env[var] = t

        return env

    if isinstance(a, ast.Return):
        if func:
            rules_no_restriction(a)
            types(a.value, env, func)
        return env

    if isinstance(a, ast.For):
        rules_no_restriction(a)
        types(a.iter, env, func)
        t_i = audits(a.iter, "types")
        if isinstance(a.target, ast.Name):
            rules_no_restriction(a.target)
            var = a.target.id
            audits(a.target, "types", int)
            if isinstance(t_i, TypeError):
                pass  # Allow the error to pass through.
            elif t_i == range:
                env[var] = int
                for a_ in a.body:
                    env = types(a_, env, func)
            else:
                audits(a.iter, "types", TypeErrorRoot("iterable must be a range"))
        return env

    if isinstance(a, ast.Expr):
        types(a.value, env, func)
        if not isinstance(audits(a.value, "rules"), SyntaxRestriction):
            rules_no_restriction(a)
        return env

    # Handle cases in which the input node is an expression.
    audits(a, "types", TypeError("type cannot be determined"))

    if isinstance(a, ast.ListComp):
        rules_no_restriction(a)
        ts = {}
        for comprehension in a.generators:
            rules_no_restriction(comprehension)
            types(comprehension.iter, env, func)
            t_c = audits(comprehension.iter, "types")
            if isinstance(comprehension.target, ast.Name):
                rules_no_restriction(comprehension.target)
                var = comprehension.target.id
                t = int
                if isinstance(t_c, TypeError):
                    pass  # Allow the error to pass through.
                elif t_c == range:
                    rules_no_restriction(comprehension.iter)
                    ts[var] = t
                else:
                    t = TypeErrorRoot("iterable must be a range value")
                    audits(comprehension.iter, "types", t)
                t = typeerror_demote(t)
                audits(comprehension.target, "types", t)
        env_ = dict(env)
        for var, t_ in ts.items():
            env_[var] = t_
        types(a.elt, env_, func)
        t_e = audits(a.elt, "types")
        if t_e is not None and not isinstance(t_e, TypeError):
            audits(a, "types", list[t_e])

    elif isinstance(a, ast.Call):
        ats = []
        kts = []
        for a_ in a.args:
            types(a_, env, func)
            ats.append(audits(a_, "types"))
        for a_ in a.keywords:
            types(a_.value, env, func)
            kts.append(audits(a_.value, "types"))

        if isinstance(a.func, ast.Attribute):
            types(a.func.value, env, func)
            t_v = audits(a.func.value, "types")
            if a.func.attr == "if_else":
                if len(a.args) == 2:
                    ts = ats
                    rules_no_restriction(a)
                    rules_no_restriction(a.func)
                    if t_v == Boolean:
                        if ts[0] in (Integer, PublicInteger, SecretInteger) and ts[
                            1
                        ] in (Integer, PublicInteger, SecretInteger):
                            t = max(ts)
                        else:
                            t = TypeErrorRoot(
                                "branches must have the same integer type"
                            )
                    elif t_v == PublicBoolean:
                        if ts[0] in (Integer, PublicInteger, SecretInteger) and ts[
                            1
                        ] in (Integer, PublicInteger, SecretInteger):
                            t = max(ts)
                        else:
                            t = TypeErrorRoot(
                                "branches must have the same integer type"
                            )
                    elif t_v == SecretBoolean:
                        if ts[0] in (Integer, PublicInteger, SecretInteger) and ts[
                            1
                        ] in (Integer, PublicInteger, SecretInteger):
                            t = max(ts)
                        else:
                            t = TypeErrorRoot(
                                "branches must have the same integer type"
                            )
                    else:
                        t = TypeErrorRoot("condition must have a boolean type")

                    audits(a, "types", t)
            elif a.func.attr == "append":
                if len(a.args) == 1:
                    rules_no_restriction(a)
                    rules_no_restriction(a.func)
                    t_i = ats[0]
                    if unify(t_v, list[t_i]):
                        audits(a, "types", type(None))
                    else:
                        audits(
                            a,
                            "types",
                            TypeErrorRoot("item type does not match list type"),
                        )

        elif isinstance(a.func, ast.Name):
            if a.func.id == "Party":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = TypeError("party requires name parameter (a string)")
                if (
                    len(a.args) == 0
                    and len(a.keywords) == 1
                    and a.keywords[0].arg == "name"
                ):
                    rules_no_restriction(a.keywords[0])
                    if audits(a.keywords[0].value, "types") == str:
                        t = Party
                elif len(a.args) == 1 and len(a.keywords) == 0:
                    rules_no_restriction(a.args[0])
                    if audits(a.args[0], "types") == str:
                        t = Party
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "Input":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = TypeError(
                    "input requires name parameter (a string) and party parameter"
                )
                if len(a.args) == 2 and len(a.keywords) == 0:
                    rules_no_restriction(a.args[0])
                    rules_no_restriction(a.args[1])
                    if (
                        audits(a.args[0], "types") == str
                        and audits(a.args[1], "types") == Party
                    ):
                        t = Input
                if (
                    len(a.args) == 1
                    and len(a.keywords) == 1
                    and a.keywords[0].arg == "party"
                ):
                    rules_no_restriction(a.args[0])
                    rules_no_restriction(a.keywords[0])
                    if (
                        audits(a.args[0], "types") == str
                        and audits(a.keywords[0].value, "types") == Party
                    ):
                        t = Input
                if (
                    len(a.args) == 0
                    and len(a.keywords) == 2
                    and a.keywords[0].arg == "name"
                    and a.keywords[1].arg == "party"
                ):
                    rules_no_restriction(a.keywords[0])
                    rules_no_restriction(a.keywords[1])
                    if (
                        audits(a.keywords[0].value, "types") == str
                        and audits(a.keywords[1].value, "types") == Party
                    ):
                        t = Input
                if (
                    len(a.args) == 0
                    and len(a.keywords) == 2
                    and a.keywords[1].arg == "name"
                    and a.keywords[0].arg == "party"
                ):
                    rules_no_restriction(a.keywords[0])
                    rules_no_restriction(a.keywords[1])
                    if (
                        audits(a.keywords[1].value, "types") == str
                        and audits(a.keywords[0].value, "types") == Party
                    ):
                        t = Input
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "Output":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = TypeError(
                    "output requires value parameter, name parameter (a string), "
                    + "and party parameter"
                )
                kwargs = {kw.arg: kw.value for kw in a.keywords}
                if len(a.args) == 0 and len(a.keywords) == 3:
                    if set(kwargs.keys()) == {"value", "name", "party"}:
                        for kw in a.keywords:
                            rules_no_restriction(kw)
                        if (
                            (
                                audits(kwargs["value"], "types")
                                in (SecretInteger, PublicInteger)
                            )
                            and audits(kwargs["name"], "types") == str
                            and audits(kwargs["party"], "types") == Party
                        ):
                            t = Output
                elif len(a.args) == 1 and len(a.keywords) == 2:
                    if set(kwargs.keys()) == {"name", "party"}:
                        for arg in a.args:
                            rules_no_restriction(arg)
                        for kw in a.keywords:
                            rules_no_restriction(kw)
                        if (
                            (
                                audits(a.args[0], "types")
                                in (SecretInteger, PublicInteger)
                            )
                            and audits(kwargs["name"], "types") == str
                            and audits(kwargs["party"], "types") == Party
                        ):
                            t = Output
                elif len(a.args) == 2 and len(a.keywords) == 1:
                    if set(kwargs.keys()) == {"party"}:
                        for arg in a.args:
                            rules_no_restriction(arg)
                        for kw in a.keywords:
                            rules_no_restriction(kw)
                        if (
                            (
                                audits(a.args[0], "types")
                                in (SecretInteger, PublicInteger)
                            )
                            and audits(a.args[1], "types") == str
                            and audits(kwargs["party"], "types") == Party
                        ):
                            t = Output
                elif len(a.args) == 3 and len(a.keywords) == 0:
                    for arg in a.args:
                        rules_no_restriction(arg)
                    for kw in a.keywords:
                        rules_no_restriction(kw)
                    if (
                        (audits(a.args[0], "types") in (SecretInteger, PublicInteger))
                        and audits(a.args[1], "types") == str
                        and audits(a.args[2], "types") == Party
                    ):
                        t = Output
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "Integer":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = Integer
                if len(a.args) != 1 or audits(a.args[0], "types") != int:
                    t = TypeError("expecting single argument (an integer)")
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "PublicInteger":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = PublicInteger
                if len(a.args) != 1 or audits(a.args[0], "types") != Input:
                    t = TypeError("expecting single argument (an input object)")
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "SecretInteger":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = SecretInteger
                if len(a.args) != 1 or audits(a.args[0], "types") != Input:
                    t = TypeError("expecting single argument (an input object)")
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "range":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = range
                if len(a.args) != 1 or audits(a.args[0], "types") != int:
                    t = TypeErrorRoot("expecting single integer argument")
                    if isinstance(audits(a.args[0], "types"), TypeError):
                        t = typeerror_demote(t)
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "str":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = str
                if len(a.args) != 1 or audits(a.args[0], "types") != int:
                    t = TypeError("expecting single integer argument")
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id == "sum":
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t = SecretInteger
                if (
                    len(a.args) != 1
                    or audits(a.args[0], "types") != list[SecretInteger]
                ):
                    t = TypeError("expecting argument of type list[SecretInteger]")
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

            elif a.func.id in env:
                rules_no_restriction(a)
                rules_no_restriction(a.func)
                t_f = env[a.func.id]
                ts = t_f.__args__[:-1]
                t = TypeErrorRoot("function arguments do not match function type")
                if len(ats) == len(ts):
                    if all(t_a == t for (t_a, t) in zip(ats, ts)):
                        t = t_f.__args__[-1]
                audits(a, "types", t)
                audits(a.func, "types", TypeInParent())

    elif isinstance(a, ast.Subscript):
        rules_no_restriction(a)
        types(a.value, env, func)
        types(a.slice, env, func)
        t_v = audits(a.value, "types")
        t_s = audits(a.slice, "types")
        if (not isinstance(t_v, TypeError)) and (not isinstance(t_s, TypeError)):
            t = TypeErrorRoot("expecting list value and integer index")
            if t_v.__name__ == "list" and t_s == int:
                if hasattr(t_v, "__args__") and len(t_v.__args__) == 1:
                    t = t_v.__args__[0]
                    audits(a, "types", t)

    elif isinstance(a, ast.List):
        rules_no_restriction(a)
        for a_ in a.elts:
            types(a_, env, func)

        ts = [audits(a_, "types") for a_ in a.elts]
        t = TypeError("lists must contain elements that are all of the same type")
        if len(set(ts)) == 0:
            t = list
        elif len(set(ts)) == 1:
            t = ts[0]
            if isinstance(t, TypeError):
                t = typeerror_demote(t)
            else:
                t = list[t]
        audits(a, "types", t)

    elif isinstance(a, ast.BoolOp):
        rules_no_restriction(a)
        ts = []
        for a_ in a.values:
            types(a_, env, func)
            ts.append(audits(a_, "types"))
        t = TypeErrorRoot("arguments must be boolean values")
        if all(t == bool for t in ts):
            t = bool
        elif any(isinstance(t, TypeError) for t in ts):
            t = typeerror_demote(t)
        audits(a, "types", t)

    elif isinstance(a, ast.BinOp):
        types(a.left, env, func)
        types(a.right, env, func)
        t_l = audits(a.left, "types")
        t_r = audits(a.right, "types")
        if isinstance(a.op, ast.Add):
            rules_no_restriction(a)
            t = TypeError("unsupported operand types")
            if t_l == int and t_r == int:
                t = int
            elif t_l == str and t_r == str:
                t = str
            else:
                t = _types_binop_mult_add_sub(t_l, t_r)
            audits(a, "types", t)
        elif isinstance(a.op, ast.Sub):
            rules_no_restriction(a)
            t = TypeError("unsupported operand types")
            if t_l == int and t_r == int:
                t = int
            else:
                t = _types_binop_mult_add_sub(t_l, t_r)
            audits(a, "types", t)
        elif isinstance(a.op, ast.Mult):
            rules_no_restriction(a)
            if t_l == int and t_r == int:
                t = int
            else:
                t = _types_binop_mult_add_sub(t_l, t_r)
            audits(a, "types", t)

    elif isinstance(a, ast.Compare):
        if len(a.comparators) != 1:
            audits(
                a,
                "rules",
                SyntaxRestriction("chained comparisons are prohibited in strict mode"),
            )
        else:
            op = a.ops[0]
            rules_no_restriction(a)
            types(a.left, env, func)
            types(a.comparators[0], env, func)
            t_l = audits(a.left, "types")
            t_r = audits(a.comparators[0], "types")
            if isinstance(op, (ast.Eq, ast.NotEq)):
                if t_l == bool and t_r == bool:
                    t = bool
                elif t_l == int and t_r == int:
                    t = bool
                elif t_l == str and t_r == str:
                    t = bool
                else:
                    t = _types_compare(t_l, t_r)
            elif isinstance(op, (ast.Lt, ast.LtE, ast.Gt, ast.GtE)):
                if t_l == int and t_r == int:
                    t = bool
                else:
                    t = _types_compare(t_l, t_r)
            audits(a, "types", t)

    elif isinstance(a, ast.UnaryOp):
        types(a.operand, env, func)
        t = audits(a.operand, "types")
        if isinstance(a.op, (ast.UAdd, ast.USub)):
            rules_no_restriction(a)
            if isinstance(t, TypeError):
                t = typeerror_demote(t)
            elif t in (int, Integer, PublicInteger, SecretInteger):
                pass
            else:
                t = TypeErrorRoot("argument must have an integer type")
            audits(a, "types", t)

        elif isinstance(a.op, ast.Not):
            rules_no_restriction(a)
            if isinstance(t, TypeError):
                t = typeerror_demote(t)
            elif t == bool:
                pass
            else:
                t = TypeErrorRoot("argument must be a boolean value")
            audits(a, "types", t)

    elif isinstance(a, ast.Name):
        rules_no_restriction(a)
        var = a.id
        audits(
            a,
            "types",
            env[var] if var in env else TypeError("name '" + var + "' is not defined"),
        )

    elif isinstance(a, ast.Constant):
        if a.value in (False, True) and str(a.value) in ("False", "True"):
            rules_no_restriction(a)
            audits(a, "types", bool)
        elif isinstance(a.value, int):
            rules_no_restriction(a)
            audits(a, "types", int)
        elif isinstance(a.value, str):
            rules_no_restriction(a)
            audits(a, "types", str)

    return env  # Always return the environment.


def strict(source: str) -> richreports.report:
    """
    Take a Python source file representing a Nada DSL program, statically
    analyze it (against the "strict" Nada DSL subset), and generate an
    interactive HTML report detailing the results.
    """
    source = source.strip()
    (atok, skips) = parse(source)
    root = atok.tree

    # Perform the static analyses.
    rules(root)
    types(root)

    # Perform the abstract execution.
    # root.body.append(ast.Expr(ast.Call(ast.Name('nada_main', ast.Load()), [], [])))
    # ast.fix_missing_locations(root)
    # exec(compile(root, path, 'exec'))

    # Add the results of the analyses to the report and ensure each line is
    # wrapped as an HTML element.
    report = richreports.report(source, line=1, column=0)
    enrich_fromaudits(report, atok)
    for i, line in enumerate(report.lines):
        if i in skips:
            report.enrich(
                (i + 1, 0),
                (i + 1, len(line) - 1),
                '<span class="rules-SyntaxError">',
                "</span>",
                skip_whitespace=True,
            )
            report.enrich(
                (i + 1, 0),
                (i + 1, len(line) - 1),
                '<span class="detail" data-detail="SyntaxError">',
                "</span>",
                skip_whitespace=True,
            )
        report.enrich((i + 1, 0), (i + 1, len(line)), "<div>", "</div>")

    return report



"""Generic type definitions for Nada."""

from typing import TypeVar

from nada_dsl.nada_types import DslType

R = TypeVar("R", bound=DslType)
T = TypeVar("T", bound=DslType)
U = TypeVar("U", bound=DslType)



"""Nada type definitions."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict, TypeAlias, Union, Type
from nada_dsl.source_ref import SourceRef


@dataclass
class Party:
    """
    Represents a party involved in the computation.

    Attributes:
        name (str): The name of the party.
    """

    name: str
    source_ref: SourceRef

    def __init__(self, name):
        self.name = name
        self.source_ref = SourceRef.back_frame()


AllTypes = Union[
    "Integer",
    "UnsignedInteger",
    "Boolean",
    "PublicInteger",
    "PublicUnsignedInteger",
    "PublicBoolean",
    "SecretInteger",
    "SecretUnsignedInteger",
    "SecretBoolean",
    "Array",
    "Tuple",
    "NTuple",
    "Object",
    "EcdsaPrivateKey",
    "EcdsaDigestMessage",
    "EcdsaSignature",
]
AllTypesType = Union[
    Type["Integer"],
    Type["UnsignedInteger"],
    Type["Boolean"],
    Type["PublicInteger"],
    Type["PublicUnsignedInteger"],
    Type["PublicBoolean"],
    Type["SecretInteger"],
    Type["SecretUnsignedInteger"],
    Type["SecretBoolean"],
    Type["Array"],
    Type["ArrayType"],
    Type["Tuple"],
    Type["NTuple"],
    Type["Object"],
    Type["EcdsaPrivateKey"],
    Type["EcdsaDigestMessage"],
    Type["EcdsaSignature"],
]
OperationType = Union[
    "Addition",
    "Subtraction",
    "Multiplication",
    "Division",
    "Modulo",
    "Power",
    "RightShift",
    "LeftShift",
    "LessThan",
    "GreaterThan",
    "LessOrEqualThan",
    "GreaterOrEqualThan",
    "Equals",
    "Input",
    "Cast",
    "Map",
    "Zip",
    "Reduce",
    "Unzip",
    "Literal",
]

""
DslTypeRepr: TypeAlias = str | Dict[str, Dict]
"""Type alias for the NadaType representation. 

This representation can be either a string ("SecretInteger") 
or a dictionary (Array{contained_types=SecretInteger, size=3}).
"""


class Mode(Enum):
    """The mode of a type"""

    CONSTANT = 1
    PUBLIC = 2
    SECRET = 3


class BaseType(Enum):
    """The base type of a scalar type."""

    BOOLEAN = 1
    INTEGER = 2
    UNSIGNED_INTEGER = 3

    def is_numeric(self) -> bool:
        """Returns true if the base type is numeric."""
        return self in (BaseType.INTEGER, BaseType.UNSIGNED_INTEGER)


# TODO: abstract?
@dataclass
class DslType:
    """Nada type class.

    This is the parent class of all nada types.

    In Nada, all the types wrap Operations. For instance, an addition between two integers
    is represented like this SecretInteger(child=Addition(...)).

    In MIR, the representation is based around operations. A MIR operation points to other
    operations and has a return type.

    To aid MIR conversion, we store in memory (`AST_OPERATIONS`) all the operations in a
    MIR-friendly format, as subclasses of ASTOperation.

    Whenever the Python interpreter constructs an instance of NadaType, it will also store
    in memory the corresponding child operation. In order to do so, the ASTOperation will
    need the type in MIR format. Which is why all instances of `NadaType` provide an implementation
    of `to_mir()`.

    """

    child: OperationType

    def __init__(self, child: OperationType):
        """NadaType default constructor

        Args:
            child (OperationType): The child operation of this Data type
        """
        self.child = child
        if self.child is not None:
            self.child.store_in_ast(self.type().to_mir())

    def __bool__(self):
        raise NotImplementedError

    @classmethod
    def is_scalar(cls) -> bool:
        """Returns True if the type is a scalar."""
        return False

    @classmethod
    def is_literal(cls) -> bool:
        """Returns True if the type is a literal."""
        return False

    @abstractmethod
    def type(self):
        """Returns a meta type for this NadaType."""
